--	Player.lua - The trail controlled by the players.
--  ----------------------------------------------------------------------------
--	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
--	apps.
--	Copyright (C) 2013  Niall Moody
--	
--	This program is free software: you can redistribute it and/or modify
--	it under the terms of the GNU General Public License as published by
--	the Free Software Foundation, either version 3 of the License, or
--	(at your option) any later version.
--
--	This program is distributed in the hope that it will be useful,
--	but WITHOUT ANY WARRANTY; without even the implied warranty of
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--	GNU General Public License for more details.
--
--	You should have received a copy of the GNU General Public License
--	along with this program.  If not, see <http://www.gnu.org/licenses/>.
--	----------------------------------------------------------------------------

--------------------------------------------------------------------------------
--	The class declaration.
class 'Player'

--------------------------------------------------------------------------------
function Player:__init(centrePos, ind)
	self.position = {}
	self.startPos = TwoFloats(centrePos)
	self.desiredPos = TwoFloats(centrePos)
	self.scale = {}
	self.desiredScale = 0.0
	self.colour = {}
	self.desiredCol = ThreeFloats(1.0)
	self.alpha = {}

	self.posStick = TwoFloats()
	self.colourStick = TwoFloats()

	--True if we're erasing.
	self.erase = false
	--The colour to erase to.
	self.eraseCol = ThreeFloats(1.0)

	--Our sounder index.
	self.index = ind
	self.pitch = ind

	self.recordPos = {}
	self.recordScale = {}
	self.recordCol = {}

	for i=1,64 do
		self.position[i] = TwoFloats(centrePos)
		self.scale[i] = 0.0
		self.colour[i] = ThreeFloats(1.0)
		self.alpha[i] = 1.0
	end
end

--------------------------------------------------------------------------------
function Player:update(sound)
	local col = 0.0

	if (self.posStick:getMagnitude() > 0.0) then
		self.desiredPos = self.desiredPos + (self.posStick * 1.0)

		if (self.desiredPos.x < 0.0) then
			self.desiredPos.x = 0.0
		elseif (self.desiredPos.x > drawingSize.x) then
			self.desiredPos.x = drawingSize.x
		end
		if (self.desiredPos.y < 0.0) then
			self.desiredPos.y = 0.0
		elseif (self.desiredPos.y > drawingSize.y) then
			self.desiredPos.y = drawingSize.y
		end
	end

	if (self.desiredPos ~= self.position[64]) then

		--Tick array along.
		for i=1,63 do
			self.position[i] = self.position[i+1]
			self.scale[i] = self.scale[i+1]
			self.colour[i] = self.colour[i+1]
		end

		--Move position towards desired position.
		self.position[64] = self.position[64] + ((self.desiredPos - self.position[64]) * 0.05)

		if (self.position[64]:getDistanceSquared(self.desiredPos) < 1.0) then
			self.position[64] = TwoFloats(self.desiredPos)
		end

		sound:setBrushPos(self.index, self.position[64]-self.position[63])
		sound:setBrushPan(self.index, (self.position[64].x/drawingSize.x))

		--Update scale.
		if (self.desiredScale ~= self.scale[64]) then
			self.scale[64] = self.scale[64] + ((self.desiredScale - self.scale[64]) * 0.1)

			if (math.abs(self.scale[64] - self.desiredScale) < 0.01) then
				self.scale[64] = self.desiredScale
			end
		end

		if self.erase then
			sound:setBrushAmp(self.index, 0.0)
		else
			sound:setBrushAmp(self.index, (self.scale[64]/16.0) * 2.0)
		end

		--Update colour.
		if (self.desiredCol ~= self.colour[64]) then
			self.colour[64] = self.colour[64] + ((self.desiredCol - self.colour[64]) * 0.05)

			if (self.colour[64]:getDistance(self.desiredCol) < 0.01) then
				self.colour[64] = ThreeFloats(self.desiredCol)
			end

			col = self.colourStick:getAngle()/(2.0 * math.pi)
			if (col < 0.0) then
				col = col + 1.0
			end
			sound:setBrushPitch(self.index, math.floor(col * 7.99))
		end

		--Update recorded data.
		self.recordPos[#self.recordPos] = self.position[64]
		self.recordScale[#self.recordScale] = self.scale[64]
		self.recordCol[#self.recordCol] = self.colour[64]
	end
end

--------------------------------------------------------------------------------
function Player:draw(d)
	d:drawConnectedLine("PaintLine",
						self.position,
						self.scale,
						self.colour,
						self.alpha)
end

--------------------------------------------------------------------------------
function Player:joyMove(action, value, sound)
	local tempCol = ThreeFloats()

	if (action == "Up") or (action == "Down") then
		if (math.abs(value) > 0.15) then
			self.posStick.y = value * 8.0
		else
			self.posStick.y = 0.0
		end
	elseif (action == "Left") or (action == "Right") then
		if (math.abs(value) > 0.15) then
			self.posStick.x = value * 8.0
		else
			self.posStick.x = 0.0
		end
	elseif (action == "TriggerLeft") then
		if (math.abs(value) > 0.05) then
			self.desiredScale = math.abs(value) * 16.0
			self.erase = false
		else
			self.desiredScale = 0.0
		end
	elseif (action == "TriggerRight") then
		if (math.abs(value) > 0.05) then
			self.desiredScale = math.abs(value) * 32.0
			self.desiredCol = self.eraseCol
			self.erase = true
		else
			self.desiredScale = 0.0
			self.erase = false
		end
	elseif (action == "SecondaryUp") or (action == "SecondaryDown") then
		if (math.abs(value) > 0.15) and not self.erase then
			self.colourStick.y = value

			tempCol.x = self.colourStick:getAngle()/(2.0 * math.pi)
			if(tempCol.x < 0.0) then
				tempCol.x = tempCol.x + 1.0
			end
			tempCol.y = 1.0
			tempCol.z = 0.5
			self.desiredCol = hslToRgb(tempCol)
		end
	elseif (action == "SecondaryLeft") or (action == "SecondaryRight") then
		if (math.abs(value) > 0.15) and not self.erase then
			self.colourStick.x = value

			tempCol.x = self.colourStick:getAngle()/(2.0 * math.pi)
			if(tempCol.x < 0.0) then
				tempCol.x = tempCol.x + 1.0
			end
			tempCol.y = 1.0
			tempCol.z = 0.5
			self.desiredCol = hslToRgb(tempCol)
		end
	end
end
