//	GamePlayState.cpp - Basic GameState which does everything via Lua.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2013  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "drawing/Drawer.h"
#include "GamePlayState.h"
#include "LuaClasses.h"
#include "LuaDrawer.h"
#include "FilePath.h"
#include "UTF8File.h"

#include "luabind/luabind.hpp"

//------------------------------------------------------------------------------
GameStateCreator<GamePlayState> GamePlayState::creator;

//------------------------------------------------------------------------------
GamePlayState::GamePlayState(Application& app, shared_ptr<Sounder>& sound):
GameState(app, sound),
luaState(0),
fade(1.0f),
fadeTween(fade, 0.0f, 0.03f),
logFile(L"NiallsAVLibLuaLog"),
luaError(false),
sounder(dynamic_pointer_cast<ShardsSounder>(sound))
{
	std::string luaFilePath;
	FilePath luaFile(FilePath::getDataFilesDir().getChild(L"scripts"));

	try
	{
		luaState = luaL_newstate();
		luaL_openlibs(luaState);
		luabind::open(luaState);

		LuaClasses::registerLuaClasses(luaState);
		LuaDrawer::registerLuaMethods(luaState);

		//Register this class' sound methods.
		luabind::module(luaState)
		[
			luabind::class_<GamePlayState>("GamePlayState")
				.def("setBrushPos", &GamePlayState::setBrushPos)
				.def("setBrushPitch", &GamePlayState::setBrushPitch)
				.def("setBrushAmp", &GamePlayState::setBrushAmp)
				.def("setBrushPan", &GamePlayState::setBrushPan)
				.def("setGranFeedback", &GamePlayState::setGranFeedback)
				.def("triggerPad", &GamePlayState::triggerPad)
				.def("pickNewPitches", &GamePlayState::pickNewPitches)
				.def("selectPitches", &GamePlayState::selectPitches)
		];

		for(int i=0;i<luaFile.getNumChildren();++i)
		{
			if(luaFile.getChild(i).getExtension() == L"lua")
			{
				UTF8File::wstringToChar(luaFile.getChild(i).getPath(), luaFilePath);
				luaL_dofile(luaState, luaFilePath.c_str());
			}
		}

		luabind::call_function<void>(luaState, "initialise", getDrawingAreaSize());
	}
	catch(luabind::error& e)
	{
		handleError(lua_tostring(luaState, -1));
	}
	catch(const std::exception &err)
	{
		handleError(err.what());
	}
}

//------------------------------------------------------------------------------
GamePlayState::~GamePlayState()
{
	if(!luaError)
		lua_close(luaState);
}

//------------------------------------------------------------------------------
void GamePlayState::update()
{
	if(!luaError)
	{
		try
		{
			luabind::call_function<void>(luaState, "update", this);
		}
		catch(luabind::error& e)
		{
			handleError(lua_tostring(luaState, -1));
		}
		catch(const std::exception &err)
		{
			handleError(err.what());
		}
	}

	if(!fadeTween.getFinished())
	{
		fadeTween.tick();
		if(fade >= 1.0f)
			quit();
	}
}

//------------------------------------------------------------------------------
void GamePlayState::draw(shared_ptr<Drawer>& d, float timestepInterp)
{
	if(!luaDrawer.getDrawer())
		luaDrawer.setDrawer(d);

	if(!luaError)
	{
		try
		{
			luabind::call_function<void>(luaState, "draw", luaDrawer);

			if(!LuaClasses::getDebugOutput().empty())
			{
				luaDrawer.drawText(LuaClasses::getDebugOutput(),
								   "DebugFont",
								   TwoFloats(32.0f, d->getDrawingAreaSize().y*0.5f),
								   d->getDrawingAreaSize().x-64.0f,
								   0,
								   1.0f,
								   ThreeFloats(1.0f, 1.0f, 1.0f),
								   1.0f);
			}
		}
		catch(luabind::error& e)
		{
			handleError(lua_tostring(luaState, -1));
		}
		catch(const std::exception &err)
		{
			handleError(err.what());
		}
	}
	else
	{
		luaDrawer.drawText(lastError,
						   "DebugFont",
						   32.0f,
						   d->getDrawingAreaSize().x-64.0f,
						   0,
						   1.0f,
						   1.0f,
						   1.0f);
	}

	if(fade > 0.0f)
		d->fillRect(0.0f, d->getDrawingAreaSize(), ThreeFloats(1.0f), fade);
}

//------------------------------------------------------------------------------
void GamePlayState::keyDown(const std::wstring& action,
						   const std::wstring& keyPressed,
						   int keyModifier)
{
	if(action == L"Escape")
	{
		fadeTween.setDelta(0.01f);
		fadeTween.reset(1.0f);
		sounder->triggerFadeOut();
	}
	else if(fade <= 0.0f)
	{
		if(keyPressed == L"F5")
		{
			LuaClasses::getDebugOutput().clear();
			nextState(L"GamePlayState");
		}
		else if(!luaError)
		{
			try
			{
				luabind::call_function<void>(luaState,
											 "keyDown",
											 getNarrowAction(action),
											 getNarrowAction(keyPressed),
											 keyModifier,
											 this);
			}
			catch(luabind::error& e)
			{
				handleError(lua_tostring(luaState, -1));
			}
			catch(const std::exception &err)
			{
				handleError(err.what());
			}
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::keyUp(const std::wstring& action,
					     const std::wstring& keyPressed,
					     int keyModifier)
{
	if(!luaError)
	{
		try
		{
			luabind::call_function<void>(luaState,
										 "keyUp",
										 getNarrowAction(action),
										 getNarrowAction(keyPressed),
										 keyModifier);
		}
		catch(luabind::error& e)
		{
			handleError(lua_tostring(luaState, -1));
		}
		catch(const std::exception &err)
		{
			handleError(err.what());
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::mouseDown(int button, const TwoFloats& pos)
{
	if(!luaError)
	{
		try
		{
			luabind::call_function<void>(luaState,
										 "mouseDown",
										 button,
										 pos);
		}
		catch(luabind::error& e)
		{
			handleError(lua_tostring(luaState, -1));
		}
		catch(const std::exception &err)
		{
			handleError(err.what());
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::mouseUp(int button, const TwoFloats& pos)
{
	if(!luaError)
	{
		try
		{
			luabind::call_function<void>(luaState,
										 "mouseUp",
										 button,
										 pos);
		}
		catch(luabind::error& e)
		{
			handleError(lua_tostring(luaState, -1));
		}
		catch(const std::exception &err)
		{
			handleError(err.what());
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::mouseMove(const TwoFloats& pos)
{
	if(!luaError)
	{
		try
		{
			luabind::call_function<void>(luaState,
										 "mouseMove",
										 pos);
		}
		catch(luabind::error& e)
		{
			handleError(lua_tostring(luaState, -1));
		}
		catch(const std::exception &err)
		{
			handleError(err.what());
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::mouseWheel(bool dir)
{
	if(!luaError)
	{
		try
		{
			luabind::call_function<void>(luaState,
										 "mouseWheel",
										 dir);
		}
		catch(luabind::error& e)
		{
			handleError(lua_tostring(luaState, -1));
		}
		catch(const std::exception &err)
		{
			handleError(err.what());
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::joyMove(const std::wstring& action,
						   float value,
						   int joystick,
						   int axis)
{
	if(!luaError)
	{
		try
		{
			luabind::call_function<void>(luaState,
										 "joyMove",
										 getNarrowAction(action),
										 value,
										 joystick,
										 axis,
										 this);
		}
		catch(luabind::error& e)
		{
			handleError(lua_tostring(luaState, -1));
		}
		catch(const std::exception &err)
		{
			handleError(err.what());
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::setBrushPos(int index, const TwoFloats& pos)
{
	int i;

	for(i=(BrushSize-1);i>0;--i)
		brushPos[index][i] = brushPos[index][i-1];
	brushPos[index][0] = pos;

	sounder->setBrushStrokesTable(index, brushPos[index]);
}

//------------------------------------------------------------------------------
void GamePlayState::setBrushPitch(int index, int val)
{
	sounder->setBrushStrokesPitch(index, val);
}

//------------------------------------------------------------------------------
void GamePlayState::setBrushAmp(int index, float val)
{
	sounder->setBrushStrokesAmp(index, val);
}

//------------------------------------------------------------------------------
void GamePlayState::setBrushPan(int index, float val)
{
	sounder->setBrushStrokesPan(index, val);
}

//------------------------------------------------------------------------------
void GamePlayState::setGranFeedback(float val)
{
	sounder->setGranFeedback(val);
}

//------------------------------------------------------------------------------
void GamePlayState::triggerPad(int val)
{
	sounder->triggerPad(val);
}

//------------------------------------------------------------------------------
void GamePlayState::pickNewPitches()
{
	sounder->pickNewPitches();
}

//------------------------------------------------------------------------------
void GamePlayState::selectPitches(int index)
{
	sounder->selectPitches(index);
}

//------------------------------------------------------------------------------
void GamePlayState::handleError(const std::string& err)
{
	lastError = err;
	logFile.write(lastError);
	luaError = true;

	if(luaState)
		lua_close(luaState);
}
