//	ShardsSounder.cpp - The game's sounder.
//  ----------------------------------------------------------------------------
//	This file is part of 'SoManyJaggedShards', an entry for 7DRL 2013.
//	Copyright (C) 2013  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ShardsSounder.h"
#include "MiscHelper.h"
#include "FilePath.h"
#include "UTF8File.h"
#include "Random.h"

#include "PastaSlow.inl"

#include <cassert>
#include <sstream>
#include <string>

//------------------------------------------------------------------------------
ShardsSounder::ShardsSounder():
pitchMultiplier(1.0f),
distortGain(0.5f),
loud(false),
currentPitchSet(-1),
fadeOut(1.0f)
{
	int i, j;
	FilePath pitchesDir(FilePath::getDataFilesDir().getChild(L"pitches"));

	for(i=0;i<pitchesDir.getNumChildren();++i)
	{
		std::wstring tempstr;
		UTF8File phil(pitchesDir.getChild(i));

		pitchSets.push_back(std::vector<float>());

		while(!phil.readLine(tempstr))
		{
			float tempf;
			std::wstringstream tempstr2;

			tempstr2 << tempstr;
			tempstr2 >> tempf;

			pitchSets.back().push_back(tempf);
		}
	}

	for(j=0;j<2;++j)
	{
		for(i=0;i<BufferSize;++i)
			granBuffers[j][i] = 0.0f;
	}

	pickNewPitches();

	for(i=0;i<4;++i)
	{
		brushStrokesPitch[i] = i;
		brushStrokes[i].setActive(true);
		brushStrokes[i].setAmplitude(2.0f);
	}
	padSynth.loadParameters(PastaSlow::numParameters, PastaSlow::parameters);
	delay.setDelay(0.66f);
	delay.setFeedback(0.66f);
	delay.setMix(0.33f);
	granulators[0].setDensity(0.66f);
	granulators[0].setDuration(2.0f);
	//granulators[0].setPitch(1.0f * powf(2.0f, 5.0f/12.0f));
	granulators[0].setPitch(0.5f);
	granulators[0].setMaxBlockSize(44100*2);
	granulators[1].setDensity(1.0f);
	granulators[1].setDuration(2.0f);
	//granulators[1].setPitch(1.0f * powf(2.0f, 7.0f/12.0f));
	granulators[1].setPitch(2.0f);
	granulators[1].setMaxBlockSize(44100*2);
	combFilters.setMix(0.33f);
	combFilters.setOutputLevel(0.05f);
}

//------------------------------------------------------------------------------
ShardsSounder::~ShardsSounder()
{

}

//------------------------------------------------------------------------------
void ShardsSounder::getAudio(float *input, float *output, int numSamples)
{
	int i;
	float dist;
	float tempf;

	assert(numSamples < (BufferSize/2));

	//Just in case.
	for(i=0;i<(numSamples*2);++i)
		output[i] = 0.0f;

	for(i=0;i<4;++i)
		brushStrokes[i].getAudio(output, numSamples);

	padSynth.getAudio(output, numSamples);

	//Process the granulators.
	for(i=0;i<(numSamples*2);++i)
	{
		granBuffers[0][i] = output[i] * 0.1f;
		granBuffers[1][i] = output[i] * 0.1f;
	}
	granulators[0].processBlock(granBuffers[0], numSamples);
	granulators[1].processBlock(granBuffers[1], numSamples);
	for(i=0;i<(numSamples*2);++i)
	{
		output[i] += granBuffers[0][i] * 2.0f;
		output[i] += granBuffers[1][i];
	}

	delay.getAudio(output, numSamples);
	combFilters.processBlock(output, numSamples);

	if(loud)
	{
		tempf = 1.0f - (fakeLog(distortGain) * 0.9f);
		dist = 1.0f+(fakeLog(distortGain)*9.0f);
		for(i=0;i<(numSamples*2);++i)
			output[i] = tanhf(output[i]*dist) * tempf;
	}

	for(i=0;i<(numSamples*2);++i)
		output[i] *= 2.0f;

	if(fadeOut < 1.0f)
	{
		for(i=0;i<numSamples;++i)
		{
			output[(i*2)] *= fadeOut;
			output[(i*2)+1] *= fadeOut;

			if(fadeOut > 0.0f)
				fadeOut -= 0.00001f;
		}
	}
}

//------------------------------------------------------------------------------
void ShardsSounder::reset()
{
	int i;
	TwoFloats emptyTrail[BrushStrokesSynth::TableSize];

	for(i=0;i>4;++i)
		brushStrokes[i].setTables(emptyTrail);
}

//------------------------------------------------------------------------------
void ShardsSounder::triggerFadeOut()
{
	fadeOut -= 0.00001f;
}

//------------------------------------------------------------------------------
void ShardsSounder::setBrushStrokesPitch(int index, int val)
{
	assert(val > -1);
	assert(val < (int)pitchSets[0].size());

	brushStrokes[index].setFrequency(pitchSets[currentPitchSet][val] * 0.25f * pitchMultiplier);
}

//------------------------------------------------------------------------------
void ShardsSounder::setBrushStrokesTable(int index, const TwoFloats *table)
{
	brushStrokes[index].setTables(table);
}

//------------------------------------------------------------------------------
void ShardsSounder::setBrushStrokesPan(int index, float val)
{
	brushStrokes[index].setPan(val);
}

//------------------------------------------------------------------------------
void ShardsSounder::setBrushStrokesAmp(int index, float val)
{
	brushStrokes[index].setAmplitude(val);
}

//------------------------------------------------------------------------------
void ShardsSounder::triggerPad(int val)
{
	padSynth.triggerNote(pitchSets[currentPitchSet][val] * 0.25f * pitchMultiplier, 0);
}

//------------------------------------------------------------------------------
void ShardsSounder::setDelay(float val)
{
	delay.setDelay(val);
}

//------------------------------------------------------------------------------
void ShardsSounder::setGranFeedback(float val)
{
	granulators[0].setFeedback(val);
	granulators[1].setFeedback(val);
}

//------------------------------------------------------------------------------
void ShardsSounder::setPitchMultiplier(float val)
{
	int i;

	pitchMultiplier = val;

	for(i=0;i<4;++i)
		brushStrokes[i].setFrequency(pitchSets[currentPitchSet][brushStrokesPitch[i]] * pitchMultiplier);

	for(i=0;i<8;++i)
		combFilters.setPitch(i, pitchSets[currentPitchSet][i] * pitchMultiplier);
}

//------------------------------------------------------------------------------
void ShardsSounder::incDistortGain()
{
	if(distortGain < 0.5f)
		distortGain += 0.01f;
}

//------------------------------------------------------------------------------
void ShardsSounder::setLoud(bool val)
{
	loud = val;
}

//------------------------------------------------------------------------------
void ShardsSounder::pickNewPitches()
{
	int i;
	int tempint = Random::getGlobal().getRandomInt(pitchSets.size());

	while(tempint == currentPitchSet)
	{
		++tempint;
		tempint %= pitchSets.size();
	}
	currentPitchSet = tempint;

	for(i=0;i<4;++i)
		setBrushStrokesPitch(i, i);

	for(i=0;i<8;++i)
		combFilters.setPitch(i, pitchSets[currentPitchSet][i] * pitchMultiplier);
}

//------------------------------------------------------------------------------
void ShardsSounder::selectPitches(int index)
{
	int i;

	currentPitchSet = index;

	for(i=0;i<4;++i)
		brushStrokes[i].setFrequency(pitchSets[currentPitchSet][brushStrokesPitch[i]] * pitchMultiplier);

	for(i=0;i<8;++i)
		combFilters.setPitch(i, pitchSets[currentPitchSet][i] * pitchMultiplier);
}
