//	SimpleSynth.h - Simple analogue-esque synth.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2010  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SIMPLESYNTH_H_
#define SIMPLESYNTH_H_

#include "MonophonicSynth.h"
#include "MoogFilter.h"

#include <cmath>

///	Simple analogue-esque synth.
class SimpleSynth : public MonophonicSynth
{
  public:
	///	Constructor.
	SimpleSynth();
	///	Destructor.
	~SimpleSynth();

	///	Generates the synth's audio.
	void getAudio(float *buffer, int numSamples);

	///	Triggers a note (note off will follow the length parameter).
	void triggerNote(float freq, int sample, float amp);
	///	Sets the synth to play indefinitely.
	void noteOn(float freq, int sample);
	///	Stops the synth playing indefinitely.
	void noteOff(int sample);

	///	Sets a parameter.
	void setParameter(const std::string& name, float value, bool direct = false);
	///	Returns the number of parameters this synth has.
	int getNumParameters() const {return 16;};
	///	Returns the value of the indexed parameter.
	float getParameter(int i) const;
	///	Returns the name of the indexed parameter.
	std::string getParameterName(int i) const;
  private:
	///	Singleton used to create the saw and square wavetables.
	class WavetableSingleton
	{
	  public:
		///	Returns the sole instance of the singleton.
		static WavetableSingleton& getInstance();

		///	Returns a pointer to the sawtooth wavetable.
		const float * const getSawTable(float freq) const
		{
			int index;
			
			if(freq <= 440.0f)
				index = 0;
			else if(freq <= (440.0f * 1.5f))
				index = 1;
			else if(freq <= (440.0f * 2.0f))
				index = 2;
			else if(freq <= (440.0f * 2.5f))
				index = 3;
			else if(freq <= (440.0f * 3.0f))
				index = 4;
			else if(freq <= (440.0f * 3.5f))
				index = 5;
			else
				index = 6;

			return sawTable[index];
		};
		///	Returns a pointer to the square wavetable.
		const float * const getSquareTable(float freq) const
		{
			int index;
			
			if(freq <= 440.0f)
				index = 0;
			else if(freq <= (440.0f * 1.5f))
				index = 1;
			else if(freq <= (440.0f * 2.0f))
				index = 2;
			else if(freq <= (440.0f * 2.5f))
				index = 3;
			else if(freq <= (440.0f * 3.0f))
				index = 4;
			else if(freq <= (440.0f * 3.5f))
				index = 5;
			else
				index = 6;

			return squareTable[index];
		};
		///	Returns the size of the wavetables.
		int getTableSize() const {return TableSize;};
#ifdef WIN32
		///	Destructor.
		~WavetableSingleton() {};
#endif
	  private:
		///	The size of the wavetables.
		enum
		{
			TableSize = 1024,
			NumMipMaps = 7
		};

		///	Constructor.
		WavetableSingleton()
		{
			int i, k;
			float index;
			unsigned int j;
			unsigned int mipSize = 128;

			//Generate the sawtooth wavetable.
			for(k=0;k<NumMipMaps;++k)
			{
				j = 1;
				while(j < mipSize)
				{
					index = 0.0f;
					for(i=0;i<TableSize;++i)
					{
						sawTable[k][i] += sinf(index * (float)j * 6.283185f) * (1.0f/(float)j) * 0.5f;
						index += (1.0f/(float)TableSize);
					}
					++j;
				}
				mipSize >>= 1; //Right shifting an int is equivalent to divding by 2.
			}

			//Generate the square wave wavetable.
			mipSize = 128;
			for(k=0;k<NumMipMaps;++k)
			{
				j = 1;
				while(j < (mipSize*2))
				{
					index = 0.0f;
					for(i=0;i<TableSize;++i)
					{
						squareTable[k][i] += sinf(index * (float)j * 6.283185f) * (1.0f/(float)j) * 0.5f;
						index += (1.0f/(float)TableSize);
					}
					j += 2;
				}
				mipSize >>= 1;
			}
		};
#ifndef WIN32
		///	Destructor.
		~WavetableSingleton() {};
#endif
		///	The sawtooth wavetable.
		float sawTable[NumMipMaps][TableSize];
		///	The square wavetable.
		float squareTable[NumMipMaps][TableSize];
	};

	//--Manipulatable Parameters--
	///	How much of the 2 waveforms is mixed in.
	float waveAmount;
	///	How much of the detuned 2nd sawtooth is mixed in.
	float detunedSawAmount;
	///	The filter's cutoff frequency.
	float filterCutoff;
	///	The filter's resonance.
	float filterRes;
	///	Whether the filter attack's an exp or a log curve.
	bool filterAttackCurve;
	///	Whether the filter release's an exp or a log curve.
	bool filterReleaseCurve;
	///	The filter's attack time.
	float filterAttackTime;
	///	The filter's release time.
	float filterReleaseTime;
	///	The filter's envelope amount (-1 -> +1).
	float filterEnvAmount;
	///	Whether the attack's an exp or a log curve.
	bool attackCurve;
	///	Whether the release's an exp or a log curve.
	bool releaseCurve;
	///	The attack time.
	float attackTime;
	///	The release time.
	float releaseTime;
	///	The time it takes to play a note.
	float noteLength;
	///	The amount of gain to apply before the distortion.
	float distGain;
	///	The octave to play notes at.
	float octave;
	///	The output level.
	float level;

	///	Per-note amplitude.
	float noteAmp;

	//--Internal Parameters--
	///	The size of the wavetables.
	int tableSize;
	///	The sawtooth wavetable.
	const float *sawTable;
	///	The square wavetable.
	const float *squareTable;
	///	Our current position within the wavetables.
	float tableIndex;
	///	The detuned saw position within the wavetable.
	float detuneIndex;

	///	The current frequency.
	float frequency;
	///	The detuned saw frequency.
	float detuneFreq;

	///	The lowpass filter.
	MoogFilter filter;

	///	Our current position within noteLength.
	float noteIndex;
	///	The length (in seconds) of a single sample.
	float increment;

	///	True if there is a note waiting to be triggered.
	bool noteWaiting;
	///	When to trigger the waiting note in the current sample block.
	int waitingSample;

	///	If true, ignore noteLength.
	bool ignoreLength;
};

#endif
